import logging
import typing

import fastapi
import keycloak
from fastapi.openapi.models import HTTPBearer as HTTPBearerModel
from fastapi.security.base import SecurityBase

from neos_common import base, error
from neos_common.authorization import token, validator
from neos_common.authorization.base import (
    AccessValidator,
    AccessValidatorDependency,
    ActionResource,
    AuthorizationDependency,
    ConfigDependency,
    IAMClientDependency,
    KeycloakClientDependency,
    SignatureValidator,
    SignatureValidatorDependency,
)
from neos_common.authorization.signer import KeyPair
from neos_common.base import ApiConfig, Config
from neos_common.client import IAMClient, KeycloakClient

logger = logging.getLogger(__name__)


def create_openapi_info(action: base.ActionBase, resource: str) -> typing.Dict[str, str]:
    """Generate openapi info for use in FastAPI routes."""
    return {
        "x-iam-action": action.value,
        "x-iam-resource": resource,
    }


class DepKeycloakClientFactory:
    """Keycloak client dependency.

    Pull the config from the app.state, instantiate a keycloak client.

    Returns:
    -------
        KeycloakClient using configured keycloak parameters.
    """

    @classmethod
    def build(cls, config_dep: typing.Type[ConfigDependency]) -> typing.Type[KeycloakClientDependency]:
        """Build a KeycloakClientDependency.

        Generate a fastapi dependency that creates a KeycloakClient.
        """

        class DepKeycloakClient:
            async def __call__(self, config: ApiConfig = fastapi.Depends(config_dep())) -> KeycloakClient:
                return KeycloakClient(
                    host=config.keycloak_host,
                    realm=config.keycloak_realm,
                    client_id=config.keycloak_client_id,
                    client_secret=config.keycloak_client_secret,
                )

        return DepKeycloakClient


class DepAuthorizationFactory:
    """Authorization dependency.

    Pull the config from the app.state and parse the token from the request.

    Returns:
    -------
        TokenData with information about current user.
    """

    @classmethod
    def build(
        cls,
        config_dep: typing.Type[ConfigDependency],
        keycloak_dep: typing.Type[KeycloakClientDependency],
        signature_dep: typing.Type[SignatureValidatorDependency],
        access_validator: typing.Type[AccessValidatorDependency],
    ) -> typing.Type[AuthorizationDependency]:
        """Build an AuthorizationDependency.

        Generate a fastapi dependency that validates incoming authorization
        headers and creates TokenData.
        """

        class DepAuthorization(SecurityBase):
            def __init__(
                self,
                action_resources: typing.Union[typing.List[ActionResource], ActionResource],
                *,
                return_allowed_resources: bool = False,
            ) -> None:
                self.model = HTTPBearerModel()
                self.scheme_name = "bearer"

                self.action_resources = (
                    [action_resources] if isinstance(action_resources, ActionResource) else action_resources
                )
                self.return_allowed_resources = return_allowed_resources

            async def __call__(
                self,
                request: fastapi.Request,
                keycloak_client: KeycloakClient = fastapi.Depends(keycloak_dep()),
                config: Config = fastapi.Depends(config_dep()),
                signature_validator: SignatureValidator = fastapi.Depends(signature_dep()),
                access_validator: AccessValidator = fastapi.Depends(access_validator()),
            ) -> token.TokenData:
                authorization = request.headers.get("Authorization", "")
                auth_type, _, credentials = authorization.partition(" ")
                if auth_type == "":
                    msg = "Missing Authorization header."
                    raise error.AuthorizationRequiredError(msg)

                actions_: typing.List[typing.Union[base.ActionBase, str]] = []
                resources_: typing.List[base.ResourceLike] = []

                for ar in self.action_resources:
                    resource_kwargs = ar.resource_extractor(request) if ar.resource_extractor else {"resource_id": None}
                    resources_.append(
                        ar.resource.generate_from_config(
                            config,
                            ar.resource_type,
                            **resource_kwargs,
                        ),
                    )
                    actions_.append(ar.action)

                if auth_type.lower() == "bearer":
                    access_token = credentials

                    logger.info(access_token)
                    try:
                        introspected_token = keycloak_client.introspect_token(access_token)
                    except keycloak.KeycloakError as e:
                        message = keycloak_client.parse_error(e)
                        logger.warning(message)
                        raise error.IdentityAccessManagerError(message) from e

                    if not introspected_token["active"]:
                        raise error.InvalidAuthorizationError

                    try:
                        user_id = introspected_token["sub"]
                    except KeyError as e:
                        msg = f"Invalid token format, {e!s}"
                        raise error.InvalidAuthorizationError(msg) from e

                    user_id, resources = (
                        await access_validator.validate(
                            user_id,
                            actions_,
                            resources_,
                            return_allowed_resources=self.return_allowed_resources,
                        )
                        if resources_ != []
                        else (user_id, [])
                    )

                    return token.TokenData(user_id, access_token, resources)

                if auth_type.lower() in ("neos4-hmac-sha256", "aws4-hmac-sha256"):
                    # TODO: handle no action/resource
                    user_id, resources = await signature_validator.validate(
                        request,
                        actions_,
                        resources_,
                        return_allowed_resources=self.return_allowed_resources,
                    )
                    return token.TokenData(user_id=str(user_id), auth_token="none", resources=resources)  # nosec: B106

                msg = "Unsupported authorization header."
                raise error.InvalidAuthorizationError(msg)

        return DepAuthorization


class DepAccessValidatorFactory:
    """Access validator dependency."""

    @classmethod
    def build(
        cls,
        iam_client_dep: typing.Type[IAMClientDependency],
    ) -> typing.Type[AccessValidatorDependency]:
        """Build an AccessValidatorDependency.

        Generate a fastapi dependency that creates an AccessValidator.
        """

        class DepAccessValidator(AccessValidatorDependency):
            async def __call__(
                self,
                iam_client: IAMClient = fastapi.Depends(iam_client_dep()),
            ) -> validator.AccessValidator:
                return validator.AccessValidator(
                    iam_client=iam_client,
                )

        return DepAccessValidator


class DepIAMClientFactory:
    @classmethod
    def build(
        cls,
        config_dep: typing.Type[ConfigDependency],
    ) -> typing.Type[IAMClientDependency]:
        """Build an IAMClientDependency.

        Generate a fastapi dependency that creates an IAMClient.
        """

        class DepIAMClient:
            async def __call__(
                self,
                config: Config = fastapi.Depends(config_dep()),
            ) -> IAMClient:
                return IAMClient(
                    host=config.access_manager_host,  # type: ignore[reportGeneralTypeIssues]
                    token=None,
                    key_pair=KeyPair(
                        config.access_key_id,  # type: ignore[reportGeneralTypeIssues]
                        config.secret_access_key,  # type: ignore[reportGeneralTypeIssues]
                        config.partition,
                    ),
                )

        return DepIAMClient



import typing
from unittest import mock

import fastapi
import pytest
from pytest import MonkeyPatch  # noqa: PT013

from neos_common import error
from neos_common.authorization.base import AuthorizationDependency
from neos_common.authorization.token import TokenData
from neos_common.base import ActionBase
from neos_common.client.iam_client import IAMClient


@pytest.fixture()
def _auth_patch_factory(  # noqa: PT005
    monkeypatch: MonkeyPatch,
) -> typing.Callable[[str, "AuthorizationDependency", typing.List[str]], None]:
    """Mock the check of the token.

    Generates :class:`TokenData` with given user.
    """

    def factory(
        user: str,
        authorization_dep: "AuthorizationDependency",
        resources: typing.Union[typing.List[str], None],
    ) -> None:
        async def __call_mock__(  # noqa: N807
            self,  # noqa: ANN001
            request: fastapi.Request,
            keycloak_client: str = "",  # noqa: ARG001
            config: str = "",
            signature_validator: str = "",  # noqa: ARG001
            access_validator: str = "",  # noqa: ARG001
        ) -> TokenData:
            if resources is None and self.action_resources != []:
                resource_kwargs = (
                    self.action_resources[0].resource_extractor(request)
                    if self.action_resources[0].resource_extractor
                    else {"resource_id": None}
                )
                resource = self.action_resources[0].resource.generate_from_config(
                    config,
                    self.action_resources[0].resource_type,
                    **resource_kwargs,
                )
                m = f"The principal <{user}> must have <{self.action_resources[0].action.value}> action for the resource <{resource.urn}>."  # noqa: E501
                raise error.InsufficientPermissionsError(m)
            return TokenData(user, request.headers.get("Authorization", ""), resources=resources or [])

        monkeypatch.setattr(authorization_dep, "__call__", __call_mock__)

    return factory


@pytest.fixture()
def iam_allow_factory(monkeypatch: "MonkeyPatch") -> typing.Callable[[str, typing.List[str]], None]:
    """IAM factory generator when user has permissions."""

    def factory(user: str, resources: typing.List[str]) -> None:
        monkeypatch.setattr(IAMClient, "validate_token", mock.AsyncMock(return_value=(user, resources)))

    return factory


@pytest.fixture()
def iam_deny_factory(monkeypatch: "MonkeyPatch") -> typing.Callable[[], None]:
    """IAM factory generator when user doesn't have permissions."""

    def factory() -> None:
        async def mock_validate_token(
            self,  # noqa: ANN001, ARG001
            principal: str,
            actions: typing.List[ActionBase],
            resources: typing.List[str],
            *,
            return_allowed_resources: bool,  # noqa: ARG001
        ) -> typing.List[str]:
            msg = f"The principal <{principal}> must have <{actions[0].value}> action for the resource <{resources[0]}>."  # noqa: E501
            raise error.InsufficientPermissionsError(msg)

        monkeypatch.setattr(IAMClient, "validate_token", mock_validate_token)

    return factory



import enum
import typing

import pydantic

from neos_common import base

ResourcePattern = pydantic.constr(pattern=base.ResourceBase.RESOURCE_PATTERN)


class Statement(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(use_enum_values=True)

    sid: str
    principal: typing.List[str]
    action: typing.List[str]
    resource: typing.List[ResourcePattern]  # type: ignore[reportGeneralTypeIssues]
    condition: typing.List[str] = pydantic.Field(default_factory=list)
    effect: base.EffectEnum

    def is_allowed(self) -> bool:
        return self.effect == base.EffectEnum.allow.value


class PriorityStatement(Statement):
    priority: int


class Statements(pydantic.BaseModel):
    statements: typing.List[Statement]


class PriorityStatements(pydantic.BaseModel):
    statements: typing.List[PriorityStatement]


class SocketRequest(pydantic.BaseModel):
    request_type: str
    data: typing.Dict[str, typing.Any]


class PrincipalType(enum.Enum):
    model_config = pydantic.ConfigDict(use_enum_values=True)

    user = "user"
    group = "group"


class Principal(pydantic.BaseModel):
    principal_id: str
    principal_type: PrincipalType


class Principals(pydantic.BaseModel):
    principals: typing.List[Principal]

    def get_principal_ids(self) -> typing.List[str]:
        return [p.principal_id for p in self.principals]


class EventPacket(pydantic.BaseModel):
    source: str
    timestamp: int  # timestamp in ms
    span_id: pydantic.UUID4
    version: str
    message: typing.Union[str, typing.Dict[str, typing.Any]]
    message_type: str

    @pydantic.field_validator("span_id")
    def span_id_to_str(cls, value: pydantic.UUID4) -> str:
        return str(value)


class EventPackets(pydantic.BaseModel):
    packets: typing.List[EventPacket]



import typing
import uuid
from dataclasses import dataclass
from enum import Enum

import fastapi

from neos_common.authorization.token import TokenData
from neos_common.base import ActionBase, ApiConfig, Config, ResourceLike
from neos_common.client import IAMClient, KeycloakClient


class SignatureValidator(typing.Protocol):
    """Define the base requirements for an object that can validate signed requests."""

    async def validate(
        self,
        request: "fastapi.Request",
        action: typing.List[typing.Union[ActionBase, str]],
        resource: typing.List[ResourceLike],
        *,
        return_allowed_resources: bool,
    ) -> typing.Tuple[uuid.UUID, typing.List[str]]:
        """Validate a request and return the associated user_id and resources."""
        ...  # pragma: no cover


class AccessValidator:
    async def validate(
        self,
        user_id: uuid.UUID,
        actions: typing.List[typing.Union[ActionBase, str]],
        resources: typing.List[ResourceLike],
        *,
        return_allowed_resources: bool = True,
    ) -> typing.Tuple[uuid.UUID, typing.List[str]]:
        ...  # pragma: no cover


class ConfigDependency(typing.Protocol):
    """Define the base requirements for a dependency that returns Config."""

    async def __call__(self, request: "fastapi.Request") -> Config:
        ...  # pragma: no cover


class SignatureValidatorDependency(typing.Protocol):
    """Define the base requirements for a dependency that returns a SignatureValidator."""

    async def __call__(self) -> SignatureValidator:
        ...  # pragma: no cover


class IAMClientDependency(typing.Protocol):
    """Define the base requirements for a dependency that returns an IAMClient."""

    async def __call__(self, config: Config) -> IAMClient:
        ...  # pragma: no cover


class AccessValidatorDependency(typing.Protocol):
    """Define the base requirements for a dependency that returns an AccessValidator."""

    async def __call__(self, iam_client: IAMClient, token: TokenData) -> AccessValidator:
        ...  # pragma: no cover


class KeycloakClientDependency(typing.Protocol):
    """Define the base requirements for a dependency that returns a KeycloakClient."""

    async def __call__(self, config: ApiConfig) -> KeycloakClient:
        ...  # pragma: no cover


@dataclass
class ActionResource:
    action: ActionBase
    resource_type: Enum
    resource: ResourceLike
    resource_extractor: typing.Union[typing.Callable[[fastapi.Request], typing.Dict[str, str]], None] = None


class AuthorizationDependency(typing.Protocol):
    """Define the base requirements for a dependency that validates authorization."""

    def __init__(
        self,
        action_resources: typing.Union[typing.List[ActionResource], ActionResource],
        *,
        return_allowed_resources: bool = False,
    ) -> None:
        ...  # pragma: no cover

    async def __call__(
        self,
        request: fastapi.Request,
        keycloak_client: KeycloakClient,
        config: Config,
        signature_validator: SignatureValidator,
        access_validator: AccessValidator,
    ) -> TokenData:
        ...  # pragma: no cover



import logging
import typing

from neos_common.authorization.signer import KeyPair
from neos_common.client.base import NeosClient

if typing.TYPE_CHECKING:
    from neos_common.authorization.token import TokenData

logger = logging.getLogger(__name__)


class RegistryClient(NeosClient):
    known_errors: typing.ClassVar[typing.Set[str]] = {
        "R100",
        "R101",
        "R110",
        "R111",
        "A200",
        "A201",
    }

    def __init__(
        self,
        host: str,
        token: "typing.Union[TokenData, None]",
        key_pair: typing.Union[KeyPair, None],
    ) -> None:
        assert token is not None or key_pair is not None

        self._token = token
        self._key_pair = key_pair

        self._host = host
        self._principals = None

    @property
    def service_name(self) -> str:
        return "Registry"

    @property
    def token(self) -> "typing.Union[TokenData, None]":
        return self._token

    @property
    def key_pair(self) -> typing.Union[KeyPair, None]:
        return self._key_pair

    @property
    def core_url(self) -> str:
        return f"{self._host}/core/{{identifier}}/announce"

    @property
    def data_product_url(self) -> str:
        return f"{self._host}/core/{{identifier}}/data_product"

    @property
    def metadata_url(self) -> str:
        return f"{self._host}/core/{{identifier}}/data_product/metadata"

    async def announce_core(self, host: str, version: str, core_identifier: str) -> None:
        r = await self._post(
            url=self.core_url.format(identifier=core_identifier),
            json={
                "host": host,
                "version": version,
            },
        )

        self.process_response(r)

    async def register_data_product(
        self,
        urn: str,
        name: str,
        metadata: dict,
        description: str,
        sanitized_name: str,
        data_product_type: str,
        engine: str,
        table: str,
        identifier: str,
        core_identifier: str,
        *,
        public: bool = True,
    ) -> None:
        json = {
            "urn": urn,
            "name": name,
            "metadata": metadata,
            "description": description,
            "sanitized_name": sanitized_name,
            "data_product_type": data_product_type,
            "engine": engine,
            "table": table,
            "identifier": identifier,
            "public": public,
        }
        logger.info(
            f"Register data product request: {json}",
        )

        r = await self._post(
            url=self.data_product_url.format(identifier=core_identifier),
            json=json,
        )

        self.process_response(r)

    async def update_data_product_metadata(
        self,
        urn: str,
        metadata: dict,
        description: str,
        core_identifier: str,
    ) -> None:
        r = await self._post(
            url=self.metadata_url.format(identifier=core_identifier),
            json={
                "urn": urn,
                "metadata": metadata,
                "description": description,
            },
        )

        self.process_response(r)

    async def deregister_data_product(self, urn: str, core_identifier: str) -> None:
        r = await self._delete(
            url=self.data_product_url.format(identifier=core_identifier),
            json={
                "urn": urn,
            },
        )

        self.process_response(r)



"""Common utils to support neos cli applications."""
import typing

from yoyo import get_backend, read_migrations
from yoyo.migrations import topological_sort


def migration_upgrade(postgres_dsn: str) -> None:
    backend = get_backend(postgres_dsn)

    migrations = read_migrations("migrations")

    with backend.lock():
        # Apply any outstanding migrations
        backend.apply_migrations(backend.to_apply(migrations))


def migration_downgrade(postgres_dsn: str, count: typing.Union[int, None] = 1) -> None:
    backend = get_backend(postgres_dsn)

    migrations = backend.to_rollback(
        read_migrations("migrations"),
    )
    if count:
        migrations = migrations[:count]

    with backend.lock():
        # Rollback selected migrations (default is latest)
        backend.rollback_migrations(migrations, force=True)


def migration_history(postgres_dsn: str) -> typing.Generator[typing.Tuple[str, str], None, None]:
    backend = get_backend(postgres_dsn)
    migrations = read_migrations("migrations")

    with backend.lock():
        migrations = migrations.__class__(topological_sort(migrations))
        applied = backend.to_rollback(migrations)

    return (("A" if m in applied else " ", m.id) for m in migrations)



import typing

from web_error.error import BadRequestException, HttpException, ServerException, UnauthorisedException


# R2XX Authorization based errors
# Auth errors AXXX
# Token errors: A0XX
# IAM errors: A1XX
class UnhandledServiceApiError(HttpException):
    pass


class ServiceApiError(HttpException):
    pass


class AuthorizationRequiredError(UnauthorisedException):
    message = "Authorization token required."
    code = "A001"


class InvalidAuthorizationError(UnauthorisedException):
    message = "Authorization token invalid."
    code = "A002"


class InsufficientPermissionsError(UnauthorisedException):
    message = "Insufficient permissions."
    code = "A100"


class InvalidResourceFormatError(BadRequestException):
    code = "A102"
    message = "Resource has invalid format."


class IdentityAccessManagerError(ServerException):
    code = "A103"
    message = "Problem with Identity Access Manager."


# Service Errors CXXX
class ServiceConnectionError(HttpException):
    code = "C001"
    status = 500

    def __init__(self, message: str, debug_message: typing.Union[str, None] = None) -> None:
        super().__init__(message, debug_message, self.code, self.status)


# SXXX Signature errors
class SignatureError(UnauthorisedException):
    code = "S001"
    message = "Signature invalid."



import logging
import typing
from uuid import UUID

import httpx

from neos_common import error
from neos_common.authorization.signer import KeyPair
from neos_common.base import ActionBase
from neos_common.client.base import NeosClient

logger = logging.getLogger(__name__)


class IAMClient(NeosClient):
    """Identity Access Manager client.

    Access to the resources based on the access statements that stores in access manager.
    """

    @property
    def service_name(self) -> str:
        return "Identity Access Manager"

    @property
    def known_errors(self) -> typing.Set[str]:
        return {"I002"}

    def process_response_with_mapping(self, response: httpx.Response) -> dict:
        try:
            return self.process_response(response)
        except (error.ServiceApiError, error.UnhandledServiceApiError) as exc:
            if exc.code == "A001":
                raise error.AuthorizationRequiredError(exc.debug_message) from exc
            if exc.code == "A002":
                raise error.InvalidAuthorizationError(exc.debug_message) from exc
            if exc.code == "A100":
                raise error.InsufficientPermissionsError(exc.debug_message) from exc
            if exc.code == "A102":
                raise error.InvalidResourceFormatError(exc.debug_message) from exc
            if exc.code == "A103":
                raise error.IdentityAccessManagerError(exc.debug_message) from exc
            if exc.code == "S001":
                raise error.SignatureError(exc.debug_message) from exc
            raise

    def __init__(
        self,
        host: str,
        token: typing.Union[str, None],
        key_pair: typing.Union[KeyPair, None],
    ) -> None:
        assert token is not None or key_pair is not None

        self._token = token
        self._key_pair = key_pair

        self._host = host
        self._principals = None

    @property
    def token(self) -> typing.Union[str, None]:
        return self._token

    @property
    def key_pair(self) -> typing.Union[KeyPair, None]:
        return self._key_pair

    @property
    def validate_token_url(self) -> str:
        return f"{self._host}/validate/token"

    @property
    def validate_signature_url(self) -> str:
        return f"{self._host}/validate/signature"

    async def validate_token(
        self,
        principal: UUID,
        actions: typing.List[ActionBase],
        resources: typing.List[str],
        *,
        return_allowed_resources: bool = False,
    ) -> typing.Tuple[UUID, typing.List[str]]:
        r = await self._get(
            url=self.validate_token_url,
            params={
                "principal_id": principal,
                "action": [action.value for action in actions],
                "resource": resources,
                "return_allowed_resources": return_allowed_resources,
            },
        )

        data = self.process_response_with_mapping(r)
        logger.info(data)
        return data["principal_id"], data["resources"]

    async def validate_signature(
        self,
        access_key: str,
        auth_schema: str,
        scope: str,
        challenge: str,
        signed_challenge: str,
        actions: typing.List[ActionBase],
        resources: typing.List[str],
        *,
        return_allowed_resources: bool = False,
    ) -> typing.Tuple[UUID, typing.List[str]]:
        r = await self._get(
            url=self.validate_signature_url,
            params={
                "access_key_id": access_key,
                "auth_schema": auth_schema,
                "scope": scope,
                "challenge": challenge,
                "signed_challenge": signed_challenge,
                "action": [action.value for action in actions],
                "resource": resources,
                "return_allowed_resources": return_allowed_resources,
            },
        )

        data = self.process_response_with_mapping(r)

        return data["principal_id"], data["resources"]



import logging
import typing
import uuid

import fastapi

from neos_common import error, schema
from neos_common.authorization import access, base, signer
from neos_common.base import ActionBase, ResourceLike
from neos_common.client import IAMClient

logger = logging.getLogger(__name__)


async def validate(
    action: typing.Union[ActionBase, str],
    resource: ResourceLike,
    principals: typing.Union[schema.Principals, typing.List[str]],
    statements: schema.Statements,
    owner: typing.Union[str, None] = None,
) -> bool:
    """Validate access using PARC.

    Validate access (authorize) by checking PARC (principals, actions, resources, conditions) and
    effect (allow/deny) of endpoint requirements against statements stored in IAM.

    Statements adhear to priorities.

    User vs Group user type.
    User is stronger.
    +----------------+-----------------+-------------------+
    | user statement | group statement | validation result |
    +----------------+-----------------+-------------------+
    | deny           | deny            | deny              |
    | deny           | allow           | deny              |
    | allow          | deny            | allow             |
    | allow          | allow           | allow             |
    +----------------+-----------------+-------------------+

    Allow vs Deny effect.
    Deny is stronger.

    User type vs Effect.
    User type is stronger.
    """
    action_str: str = action.value if isinstance(action, ActionBase) else action

    error_message = "The principal <{}> must have <{}> action for the resource <{}>.".format(
        owner or principals,
        action_str,
        resource.urn,
    )

    def raise_on_empty(statements: schema.Statements) -> None:
        if not statements.statements:
            raise error.InsufficientPermissionsError(error_message)

    filtered_statements = access.filter_by_principals(statements, principals)
    raise_on_empty(filtered_statements)

    filtered_statements = access.filter_by_action(filtered_statements, action)
    raise_on_empty(filtered_statements)

    filtered_statements = access.filter_by_resource(filtered_statements, resource)
    raise_on_empty(filtered_statements)

    ordered_priority_statements = access.order_by_priority_ascending(filtered_statements, principals)

    if not ordered_priority_statements.statements[-1].is_allowed():
        raise error.InsufficientPermissionsError(error_message)

    return True


class AccessValidator(base.AccessValidator):
    def __init__(self, iam_client: IAMClient) -> None:
        self._iam_client = iam_client

    async def validate(
        self,
        user_id: uuid.UUID,
        actions: typing.List[ActionBase],
        resources: typing.List[ResourceLike],
        *,
        return_allowed_resources: bool = False,
    ) -> typing.Tuple[uuid.UUID, typing.List[str]]:
        return await self._iam_client.validate_token(
            principal=user_id,
            actions=actions,
            resources=[resource.urn for resource in resources],
            return_allowed_resources=return_allowed_resources,
        )


class SignatureValidator(base.SignatureValidator):
    def __init__(self, iam_client: IAMClient) -> None:
        self._iam_client = iam_client

    async def validate(
        self,
        request: fastapi.Request,
        actions: typing.List[ActionBase],
        resources: typing.List[ResourceLike],
        *,
        return_allowed_resources: bool = False,
    ) -> typing.Tuple[uuid.UUID, typing.List[str]]:
        validator = signer.Validator()

        payload = await request.body()

        auth_type, access_key_id, scope, challenge, signature = validator.challenge_v4(
            request.method,
            request.url,
            request.headers,
            payload,
        )

        return await self._iam_client.validate_signature(
            access_key_id,
            auth_type.split("-")[0],
            scope,
            challenge,
            signature,
            actions,
            [resource.urn for resource in resources],
            return_allowed_resources=return_allowed_resources,
        )



"""Based on https://github.com/dmontagu/fastapi-utils.

Which is based on https://github.com/steinnes/timing-asgi.git

Bringing in house to support pydantic v2 and fastapi-utils has not been updated in 4 months at 07/2023.

The middleware from this module is intended for use during both development and production,
but only reports timing data at the granularity of individual endpoint calls.

For more detailed performance investigations (during development only, due to added overhead),
consider using the coroutine-aware profiling library `yappi`.
"""
from __future__ import annotations

import resource
import time
from typing import TYPE_CHECKING, Any

from starlette.routing import Match, Mount

if TYPE_CHECKING:
    from collections.abc import Callable

    from fastapi import FastAPI
    from starlette.middleware.base import RequestResponseEndpoint
    from starlette.requests import Request
    from starlette.responses import Response
    from starlette.types import Scope

TIMER_ATTRIBUTE = "__fastapi_utils_timer__"


def add_timing_middleware(
    app: FastAPI,
    record: Callable[[str], None] | None = None,
    prefix: str = "",
    exclude: str | None = None,
) -> None:
    """Adds a middleware to the provided `app` that records timing metrics using the provided `record` callable.

    Typically `record` would be something like `logger.info` for a `logging.Logger` instance.

    The provided `prefix` is used when generating route names.

    If `exclude` is provided, timings for any routes containing `exclude`
    as an exact substring of the generated metric name will not be logged.
    This provides an easy way to disable logging for routes

    The `exclude` will probably be replaced by a regex match at some point in the future. (PR welcome!)
    """
    metric_namer = _MetricNamer(prefix=prefix, app=app)

    @app.middleware("http")
    async def timing_middleware(request: Request, call_next: RequestResponseEndpoint) -> Response:
        metric_name = metric_namer(request.scope)
        with _TimingStats(metric_name, record=record, exclude=exclude) as timer:
            setattr(request.state, TIMER_ATTRIBUTE, timer)
            return await call_next(request)


def record_timing(request: Request, note: str | None = None) -> None:
    """Call this function at any point that you want to display elapsed time during the handling of a single request.

    This can help profile which piece of a request is causing a performance bottleneck.

    Note that for this function to succeed, the request should have been generated by a FastAPI app
    that has had timing middleware added using the `fastapi_utils.timing.add_timing_middleware` function.
    """
    timer = getattr(request.state, TIMER_ATTRIBUTE, None)
    if timer is not None:
        assert isinstance(timer, _TimingStats)
        timer.emit(note)
    else:
        msg = "No timer present on request"
        raise ValueError(msg)


class _TimingStats:
    """This class tracks and records endpoint timing data.

    Should be used as a context manager; on exit, timing stats will be emitted.

    name:
        The name to include with the recorded timing data
    record:
        The callable to call on generated messages. Defaults to `print`, but typically
        something like `logger.info` for a `logging.Logger` instance would be preferable.
    exclude:
        An optional string; if it is not None and occurs inside `name`, no stats will be emitted
    """

    def __init__(
        self,
        name: str | None = None,
        record: Callable[[str], None] | None = None,
        exclude: str | None = None,
    ) -> None:
        self.name = name
        self.record = record or print

        self.start_time: float = 0
        self.start_cpu_time: float = 0
        self.end_cpu_time: float = 0
        self.end_time: float = 0
        self.silent: bool = False

        if self.name is not None and exclude is not None and (exclude in self.name):
            self.silent = True

    def start(self) -> None:
        self.start_time = time.time()
        self.start_cpu_time = _get_cpu_time()

    def take_split(self) -> None:
        self.end_time = time.time()
        self.end_cpu_time = _get_cpu_time()

    @property
    def time(self) -> float:
        return self.end_time - self.start_time

    @property
    def cpu_time(self) -> float:
        return self.end_cpu_time - self.start_cpu_time

    def __enter__(self) -> _TimingStats:
        self.start()
        return self

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:  # noqa: ANN401
        self.emit()

    def emit(self, note: str | None = None) -> None:
        """Emit timing information, optionally including a specified note."""
        if not self.silent:
            self.take_split()
            cpu_ms = 1000 * self.cpu_time
            wall_ms = 1000 * self.time
            message = f"TIMING: Wall: {wall_ms:6.1f}ms | CPU: {cpu_ms:6.1f}ms | {self.name}"
            if note is not None:
                message += f" ({note})"
            self.record(message)


class _MetricNamer:
    """This class generates the route "name" used when logging timing records.

    If the route has `endpoint` and `name` attributes, the endpoint's module and route's name will be used
    (along with an optional prefix that can be used, e.g., to distinguish between multiple mounted ASGI apps).

    By default, in FastAPI the route name is the `__name__` of the route's function (or type if it is a callable class
    instance).

    For example, with prefix == "custom", a function defined in the module `app.crud` with name `read_item`
    would get name `custom.app.crud.read_item`. If the empty string were used as the prefix, the result would be
    just "app.crud.read_item".

    For starlette.routing.Mount instances, the name of the type of `route.app` is used in a slightly different format.

    For other routes missing either an endpoint or name, the raw route path is included in the generated name.
    """

    def __init__(self, prefix: str, app: FastAPI) -> None:
        if prefix:
            prefix += "."
        self.prefix = prefix
        self.app = app

    def __call__(self, scope: Scope) -> str:
        """Generates the actual name to use when logging timing metrics for a specified ASGI Scope."""
        route = None
        for r in self.app.router.routes:
            if r.matches(scope)[0] == Match.FULL:
                route = r
                break
        if hasattr(route, "endpoint") and hasattr(route, "name"):
            name = f"{self.prefix}{route.endpoint.__module__}.{route.name}"
        elif isinstance(route, Mount):
            name = f"{type(route.app).__name__}<{route.name!r}>"
        else:
            name = str(f"<Path: {scope['path']}>")
        return name


def _get_cpu_time() -> float:
    """Generates the cpu time to report.

    Adds the user and system time, following the implementation from timing-asgi.
    """
    resources = resource.getrusage(resource.RUSAGE_SELF)
    # add up user time (ru_utime) and system time (ru_stime)
    return resources[0] + resources[1]



import json
import logging
import socket
import socketserver
import time
import typing

import pydantic
import pydantic.error_wrappers

from neos_common import error, schema

if typing.TYPE_CHECKING:
    Loc = typing.Tuple[typing.Union[int, str], ...]

    class _ErrorDictRequired(typing.TypedDict):
        loc: Loc
        msg: str
        type: str

    class ErrorDict(_ErrorDictRequired, total=False):
        """Type hint for pydantic.error_wrappers.ErrorDict.

        Stolen from type hints in pydantic.error_wrappers
        """

        ctx: typing.Dict[str, typing.Any]


logger = logging.getLogger(__name__)


Socket = socket.socket


class SocketException(Exception):  # noqa: N818
    """Base class for all socket exceptions."""

    status = "unhandled"

    def __init__(self, reason: str, message: str, details: typing.Union[typing.List["ErrorDict"], None]) -> None:
        """Socket exception initiator.

        Args:
        ----
        reason: error reason i.e. "NOT_FOUND"
        message: error message
        details: optional list of pydantic validation errors
        """
        self.reason = reason
        self.message = message
        self.details = details

    def __iter__(self) -> typing.Iterator[typing.Tuple[str, typing.Union[bool, str, typing.List["ErrorDict"], None]]]:
        """Return an iterator to support `dict(self)`."""
        return iter(
            [
                ("ok", False),
                ("status", self.status),
                ("reason", self.reason),
                ("message", self.message),
                ("details", self.details),
            ],
        )


class ValidationError(SocketException):
    status = "validation_error"


def encode(obj: dict, encoder: typing.Union[typing.Type[json.JSONEncoder], None] = None) -> bytes:
    """JSON encode message object for sending via socket."""
    new = json.dumps(obj, cls=encoder)
    return new.encode()


def decode(obj: typing.Union[bytes, None]) -> typing.Union[typing.Dict[str, typing.Any], None]:
    """Decode message received via socket."""
    if obj is None:
        return obj

    data: typing.Dict[str, typing.Any] = json.loads(obj.decode())
    return data


BYTE_COUNT = 4


def send_msg(sock: Socket, msg: bytes) -> None:
    """Send a message via the socket."""
    msg = add_msg_header(msg)
    sock.sendall(msg)


def recv_msg(sock: Socket) -> typing.Union[bytearray, None]:
    """Receive a message via the socket."""
    # start by getting the header
    # (which is an int of length `BYTE_COUNT`).
    # The header tells the message size in bytes.
    raw_msglen = recvall(sock, BYTE_COUNT)
    if not raw_msglen:
        return None
    # Then retrieve a message of length `raw_msglen`
    # this will be the actual message
    msglen = len_frombytes(raw_msglen)
    return recvall(sock, msglen)


def recvall(sock: Socket, length: int) -> typing.Union[bytearray, None]:
    """Get a message of a certain length from the socket stream."""
    data = bytearray()
    while len(data) < length:
        packet = sock.recv(length - len(data))
        if not packet:
            return None
        data.extend(packet)
    return data


def add_msg_header(msg: bytes) -> bytes:
    """Prepend message length header onto message.

    Args:
    ----
    msg: message bytestring

    Returns:
    -------
    new bytestring with original message length prepended.
    """
    return len_inbytes(msg) + msg


def len_inbytes(msg: bytes) -> bytes:
    """Retrieve length of message as a bytestring."""
    return len(msg).to_bytes(BYTE_COUNT, byteorder="big")


def len_frombytes(bmsg: bytes) -> int:
    """Extract length of message from a bytestring header."""
    return int.from_bytes(bmsg, byteorder="big")


class TCPHandler(socketserver.BaseRequestHandler):
    """Base handler for TCP socket server.

    When a message is received:
        * decode it
        * format it into an instance of `cls.Request`
        * pass it to the handler defined for `request.request_type` in `cls.request_mapping`

    Define the mapping between Request.request_typpe and handlers in `cls.request_mapping`.
    Override the default SocketRequest schema via `cls.Request`
    """

    request_mapping: typing.ClassVar[dict] = {}
    Request = schema.SocketRequest

    def handle(self) -> None:
        """Handle an incoming request.

        Handle issues decoding, processing and validating request messages, on
        error response to the client, with error details.

        On success pass the validated request message to the appropriate
        handler.
        """
        try:
            data = decode(recv_msg(self.request))
        except json.decoder.JSONDecodeError as e:
            logger.exception("Decode error processing request.")
            exc = ValidationError(
                reason="DecodeError",
                message=str(e),
                details=None,
            )
            send_msg(self.request, encode(dict(exc)))
        except Exception as e:
            logger.exception("Exception processing request.")
            exc = SocketException(
                reason="Unhandled",
                message=str(e),
                details=None,
            )
            send_msg(self.request, encode(dict(exc)))
        else:
            if data is None:
                return

            try:
                request = self.Request(**data)
                request_type = request.request_type

                RequestDataCls = self.request_mapping.get(request_type, dict)  # noqa: N806
                request_data = RequestDataCls(**request.data)
            except pydantic.ValidationError as e:
                logger.exception("Validation error processing request data.")
                exc = ValidationError(
                    reason="ValidationError",
                    message="Unable to parse message.",
                    details=e.errors(),
                )
                send_msg(self.request, encode(dict(exc)))
            except Exception as e:
                logger.exception("Exception processing request data.")
                exc = SocketException(
                    reason="Unhandled",
                    message=str(e),
                    details=None,
                )
                send_msg(self.request, encode(dict(exc)))
            else:
                self._handle(self.request, request_type, request_data)

    def _handle(self, sock: Socket, request_type: str, request_data: typing.Union[pydantic.BaseModel, dict]) -> None:
        ...


class TCPClient:
    """Base implementation for socket service client."""

    def __init__(self, host: str, port: int, timeout: int = 10, wait: int = 10) -> None:
        """TCPClient instantiator.

        Args:
        ----
        host: socket service host name
        port: socket service port
        timeout: number of seconds to wait for a connection
        wait: number of seconds to wait for a response from service
        """
        self.addr = (host, port)
        self.timeout = timeout
        self.wait = wait

    def send_request(self, request: dict) -> dict:
        """Send a request to socket service.

        Handle socket connection errors, and response timeouts.
        """
        t0 = time.time()
        try:
            sock = socket.create_connection(
                self.addr,
                timeout=self.timeout,
            )
        except socket.gaierror as e:
            raise error.ServiceConnectionError(
                message="Unable to connect to service.",
                debug_message=f"{self.addr}: {e!s}",
            ) from e

        send_msg(sock, encode(request))
        t0 = time.time()
        response = None
        try:
            while not response:
                try:
                    response = decode(recv_msg(sock))
                except Exception as e:  # noqa: BLE001
                    raise error.ServiceConnectionError(
                        message="Unable to process request to service.",
                        debug_message=f"{self.addr}: {e.__class__.__name__}({e!s})",
                    ) from e
                if time.time() - t0 > self.wait:
                    msg = "Timed out waiting for resp."
                    raise TimeoutError(msg)
        finally:
            sock.close()
        return response



import json
import logging
import typing
from datetime import datetime, timezone
from enum import Enum

import httpx
import requests

from neos_common import error
from neos_common.authorization import signer

RequestType = typing.TypeVar("RequestType", httpx.Request, requests.Request)

logger = logging.getLogger(__name__)


BAD_REQUEST_CODE = 400
REQUEST_VALIDATION_CODE = 422
RATE_LIMIT_CODE = 429


class Method(Enum):
    """HTTP request methods."""

    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"


async def log_request(request: httpx.Request) -> None:
    """Event hook for httpx events to log requests."""
    logger.info(
        f"[Request] {request.method.upper()} {request.url}",
    )


class NeosBearerClientAuth:
    def __init__(self, token: str) -> None:
        self.token = token

    def __call__(self, request: RequestType) -> RequestType:
        request.headers["Authorization"] = f"Bearer {self.token}"
        return request

    def __eq__(self, other: object) -> bool:  # noqa: D105
        return isinstance(other, NeosBearerClientAuth) and other.token == self.token


class NeosSignatureClientAuth:
    """HTTP client authentication class for NEOS rest services."""

    def __init__(self, key_pair: signer.KeyPair, service: str) -> None:
        self.key_pair = key_pair
        self.service = service

    def __call__(self, request: RequestType) -> RequestType:

        dt = datetime.now(tz=timezone.utc)
        request.headers["x-neos-date"] = signer.to_amz_date(dt)
        request.headers["host"] = request.url.netloc.decode("utf-8")

        body = request.content.decode("utf-8")
        if body:
            request.headers["Content-Length"] = str(len(body))

        s = signer.Signer("NEOS4-HMAC-SHA256")
        s.sign_v4(
            self.service,
            request.method,
            request.url,
            self.key_pair.partition,
            request.headers,
            self.key_pair,
            body,
            dt,
        )

        return request

    def __eq__(self, other: object) -> bool:  # noqa: D105
        return isinstance(other, NeosSignatureClientAuth) and other.key_pair == self.key_pair


class NeosClient(typing.Protocol):
    """Base class for HTTP client implementations for NEOS rest services."""

    unhandled_error_class = error.UnhandledServiceApiError
    handled_error_class = error.ServiceApiError

    @property
    def token(self) -> typing.Union[str, None]:
        ...

    @property
    def key_pair(self) -> typing.Union[signer.KeyPair, None]:
        ...

    @property
    def known_errors(self) -> typing.Set[str]:
        ...  # pragma: no cover

    @property
    def service_name(self) -> str:
        ...  # pragma: no cover

    async def _request(
        self,
        url: str,
        method: Method,
        params: typing.Union[dict, None] = None,
        headers: typing.Union[dict, None] = None,
        json: typing.Union[dict, None] = None,
        *,
        verify: bool = True,
        **kwargs,
    ) -> httpx.Response:
        if self.key_pair is not None:
            auth = NeosSignatureClientAuth(
                self.key_pair,
                self.service_name,
            )
        elif self.token:
            auth = NeosBearerClientAuth(self.token)
        else:
            auth = None

        async with httpx.AsyncClient(event_hooks={"request": [log_request]}, verify=verify) as client:
            try:
                r = await client.request(
                    url=url,
                    method=method.value,
                    params=params,
                    json=json,
                    headers=headers,
                    auth=auth,
                    **kwargs,
                )
            except httpx.ConnectError as e:
                raise error.ServiceConnectionError(
                    message=f"Error connecting to {self.service_name} service.",
                    debug_message=str(e),
                ) from e

        return r

    async def _get(
        self,
        url: str,
        params: typing.Union[dict, None] = None,
        headers: typing.Union[dict, None] = None,
    ) -> httpx.Response:
        return await self._request(
            url=url,
            method=Method.GET,
            params=params,
            headers=headers,
        )

    async def _post(
        self,
        url: str,
        json: typing.Union[dict, None] = None,
        headers: typing.Union[dict, None] = None,
    ) -> httpx.Response:
        return await self._request(
            url=url,
            method=Method.POST,
            json=json,
            headers=headers,
        )

    async def _put(
        self,
        url: str,
        json: typing.Union[dict, None] = None,
        headers: typing.Union[dict, None] = None,
    ) -> httpx.Response:
        return await self._request(
            url=url,
            method=Method.PUT,
            json=json,
            headers=headers,
        )

    async def _delete(
        self,
        url: str,
        json: typing.Union[dict, None] = None,
        headers: typing.Union[dict, None] = None,
    ) -> httpx.Response:
        return await self._request(
            url=url,
            method=Method.DELETE,
            json=json,
            headers=headers,
        )

    def process_response(self, response: httpx.Response) -> dict:
        try:
            data = response.json()
        except json.JSONDecodeError as exc:
            logger.error(response.content)  # noqa: TRY400
            raise self.unhandled_error_class(
                message=f"Invalid {self.service_name} api JSON format response.",
                status=response.status_code,
                code=None,
                debug_message=f"{exc.msg}: {exc.pos}",
            ) from exc
        if response.status_code >= BAD_REQUEST_CODE:
            logger.info(data)
            try:
                debug_message = None
                if data["code"] is not None:
                    debug_message = data.get("debug_message", data["message"])

                exc = self.unhandled_error_class(
                    message=f"Unhandled {self.service_name} api error response.",
                    status=response.status_code,
                    code=data["code"],
                    debug_message=debug_message,
                )
            except KeyError:
                exc = self.unhandled_error_class(
                    message=f"Unhandled {self.service_name} api response.",
                    status=response.status_code,
                    code=None,
                    debug_message=None,
                )
                if response.status_code == RATE_LIMIT_CODE:
                    exc = self.unhandled_error_class(
                        message=f"{self.service_name} api rate limit error.",
                        status=response.status_code,
                        code=None,
                        debug_message=None,
                    )

                raise exc

            if response.status_code == REQUEST_VALIDATION_CODE:
                exc = self.handled_error_class(
                    message=f"{self.service_name} request validation error.",
                    debug_message=data["debug_message"],
                    code=data["code"],
                    status=response.status_code,
                )
            elif data["code"] in self.known_errors:
                exc = self.handled_error_class(
                    message=data["message"],
                    debug_message=data["debug_message"],
                    code=data["code"],
                    status=response.status_code,
                )

            raise exc

        return data



import hashlib
import hmac
import logging
import re
import typing
import urllib.parse
from collections import OrderedDict
from datetime import datetime, timezone

import httpx
import multidict

from neos_common import error

logger = logging.getLogger(__name__)


class KeyPair(typing.NamedTuple):
    access_key_id: "str"
    secret_access_key: "str"
    partition: "str"


_MULTI_SPACE_REGEX = re.compile(r"( +)")


def _quote(
    resource: str,
    safe: str = "/",
    encoding: typing.Union[str, None] = None,
    errors: typing.Union[str, None] = None,
) -> str:
    return urllib.parse.quote(
        resource,
        safe=safe,
        encoding=encoding,
        errors=errors,
    ).replace("%7E", "~")


def _to_utc(value: datetime) -> datetime:
    """Convert to UTC time if value is not naive."""
    return value.astimezone(timezone.utc).replace(tzinfo=None) if value.tzinfo else value


def to_amz_date(value: datetime) -> str:
    """Format datetime into AMZ date formatted string."""
    return _to_utc(value).strftime("%Y%m%dT%H%M%SZ")


def to_signer_date(value: datetime) -> str:
    """Format datetime into SignatureV4 date formatted string."""
    return _to_utc(value).strftime("%Y%m%d")


def sha256_hash(data: typing.Union[bytes, str, None]) -> str:
    """Compute SHA-256 of data and return hash as hex encoded value."""
    data = data or b""
    data_ = data.encode() if isinstance(data, str) else data
    hasher = hashlib.sha256()
    hasher.update(data_)
    sha256sum = hasher.hexdigest()
    return sha256sum.decode() if isinstance(sha256sum, bytes) else sha256sum


def _hmac_hash(
    key: typing.Union[bytes, bytearray],
    data: bytes,
    *,
    hexdigest: bool = False,
) -> typing.Union[str, bytes]:
    """Return HMacSHA256 digest of given key and data."""
    hasher = hmac.new(key, data, hashlib.sha256)
    return hasher.hexdigest() if hexdigest else hasher.digest()


def _generate_canonical_headers(headers: multidict.CIMultiDict) -> typing.Tuple[str, str]:
    """Get canonical headers."""
    canonical_headers = {}
    for key, values in headers.items():
        key_ = key.lower()
        if key_ not in ("authorization", "user-agent", "accept", "accept-encoding", "connection"):
            values_ = values if isinstance(values, (list, tuple)) else [values]
            canonical_headers[key_] = ",".join([_MULTI_SPACE_REGEX.sub(" ", value) for value in values_])

    canonical_headers = OrderedDict(sorted(canonical_headers.items()))
    signed_headers = ";".join(canonical_headers.keys())
    canonical_headers = "\n".join(
        [f"{key}:{value}" for key, value in canonical_headers.items()],
    )
    return canonical_headers, signed_headers


def _recreate_canonical_headers(headers: multidict.CIMultiDict, signed_headers: str) -> str:
    """Get canonical headers from SignedHeaders."""
    signed_headers_ = signed_headers.split(";")
    canonical_headers = {}
    for key, values in headers.items():
        key_ = key.lower()
        if key_ in signed_headers_:
            values_ = values if isinstance(values, (list, tuple)) else [values]
            canonical_headers[key_] = ",".join([_MULTI_SPACE_REGEX.sub(" ", value) for value in values_])

    canonical_headers = dict(sorted(canonical_headers.items()))
    return "\n".join(
        [f"{key}:{value}" for key, value in canonical_headers.items()],
    )


def _generate_canonical_query_string(query: typing.Union[bytes, str]) -> str:
    """Get canonical query string."""
    query = query or ""
    query_: str = query.decode() if isinstance(query, bytes) else query
    return "&".join(
        [
            "=".join(pair)
            for pair in sorted(
                [params.split("=") for params in query_.split("&")],
            )
        ],
    )


def _generate_signing_key(
    schema: str,
    secret_key: str,
    date: str,
    partition: str,
    service_name: str,
) -> typing.Union[str, bytes]:
    """Get signing key."""
    logger.debug(f"signing_key schema: {schema}")
    date_key = _hmac_hash(
        (schema + secret_key).encode(),
        date.encode(),
    )
    date_region_key = _hmac_hash(date_key, partition.encode())
    date_region_service_key = _hmac_hash(
        date_region_key,
        service_name.encode(),
    )
    return _hmac_hash(date_region_service_key, f"{schema.lower()}_request".encode())


def _generate_signature(signing_key: typing.Union[str, bytes], string_to_sign: str) -> typing.Union[str, bytes]:
    """Get signature."""
    return _hmac_hash(signing_key, string_to_sign.encode(), hexdigest=True)


class Signer:
    def __init__(self, auth_type: str) -> None:
        self.auth_type = auth_type
        self.auth_schema = auth_type.split("-")[0]
        self.header_prefix = "x-neos" if self.auth_schema == "NEOS4" else "x-amz"

    @staticmethod
    def _generate_scope(schema: str, date: datetime, partition: str, service_name: str) -> str:
        """Get scope string."""
        return f"{to_signer_date(date)}/{partition}/{service_name}/{schema.lower()}_request"

    @staticmethod
    def _generate_string_to_sign(auth_type: str, date: datetime, scope: str, canonical_request_hash: str) -> str:
        """Get string-to-sign."""
        return f"{auth_type}\n{to_amz_date(date)}\n{scope}\n{canonical_request_hash}"

    @staticmethod
    def _generate_canonical_request_hash(
        method: str,
        url: httpx.URL,
        headers: multidict.CIMultiDict,
        content_sha256: str,
    ) -> typing.Tuple[str, str]:
        """Get canonical request hash.

        https://docs.aws.amazon.com/IAM/latest/UserGuide/create-signed-request.html

        Create a canonical request by concatenating the following strings, separated by newline characters.
        This helps ensure that the signature that you calculate and the signature that the server calculates can match.

            HTTPMethod
            CanonicalUri
            CanonicalQueryString
            CanonicalHeaders
            SignedHeaders
            HashedPayload

        HTTPMethod -
            The HTTP method.

        CanonicalUri -
            The URI-encoded version of the absolute path component URL
            (everything between the host and the question mark character (?) that starts the query string parameters).
            If the absolute path is empty, use a forward slash character (/).

        CanonicalQueryString -
            The URL-encoded query string parameters, separated by ampersands (&). Percent-encode reserved characters,
            including the space character. Encode names and values separately.
            If there are empty parameters, append the equals sign to the parameter name before encoding.
            After encoding, sort the parameters alphabetically by key name.
            If there is no query string, use an empty string ("").

        CanonicalHeaders -
            The request headers, that will be signed, and their values, separated by newline characters.
            Header names must use lowercase characters, must appear in alphabetical order,
            and must be followed by a colon (:). For the values, trim any leading or trailing spaces,
            convert sequential spaces to a single space, and separate the values for a multi-value header using commas.
            You must include the host header (HTTP/1.1) or the :authority header (HTTP/2),
            and any x-amz-* headers in the signature.
            You can optionally include other standard headers in the signature, such as content-type.

        SignedHeaders -
            The list of headers that you included in CanonicalHeaders, separated by semicolons (;).
            This indicates which headers are part of the signing process.
            Header names must use lowercase characters and must appear in alphabetical order.

        HashedPayload -
            A string created using the payload in the body of the HTTP request as input to a hash function.
            This string uses lowercase hexadecimal characters. If the payload is empty,
            use an empty string as the input to the hash function.
        """
        canonical_headers, signed_headers = _generate_canonical_headers(headers)
        canonical_query_string = _generate_canonical_query_string(url.query)

        path = _quote(url.path or "/")
        # CanonicalRequest =
        #   HTTPRequestMethod + '\n' +
        #   CanonicalURI + '\n' +
        #   CanonicalQueryString + '\n' +
        #   CanonicalHeaders + '\n\n' +
        #   SignedHeaders + '\n' +
        canonical_request = (
            f"{method}\n"
            f"{path}\n"
            f"{canonical_query_string}\n"
            f"{canonical_headers}\n\n"
            f"{signed_headers}\n"
            f"{content_sha256}"
        )
        logger.debug(f"canonical_request: {canonical_request}")

        return sha256_hash(canonical_request), signed_headers

    @staticmethod
    def _generate_authorization(
        auth_type: str,
        access_key: str,
        scope: str,
        signed_headers: str,
        signature: typing.Union[str, bytes],
    ) -> str:
        """Get authorization."""
        return f"{auth_type} Credential={access_key}/{scope}, SignedHeaders={signed_headers}, Signature={signature}"

    def sign_v4(
        self,
        service_name: str,
        method: str,
        url: httpx.URL,
        partition: str,
        headers: multidict.CIMultiDict,
        credentials: KeyPair,
        content: typing.Union[bytes, None],
        date: datetime,
    ) -> multidict.CIMultiDict:
        logger.debug(f"url: {url}")
        logger.debug(f"headers: {headers}")
        content_sha256 = sha256_hash(content) if url.scheme == "http" else "UNSIGNED-PAYLOAD"
        content_header = f"{self.header_prefix}-content-sha256"
        if content_header not in headers:
            headers[content_header] = content_sha256
        scope = self._generate_scope(self.auth_schema, date, partition, service_name)
        canonical_request_hash, signed_headers = self._generate_canonical_request_hash(
            method,
            url,
            headers,
            content_sha256,
        )
        string_to_sign = self._generate_string_to_sign(self.auth_type, date, scope, canonical_request_hash)
        logger.debug(f"scope: {scope}")
        logger.debug(f"string_to_sign: {string_to_sign}")
        signing_key = _generate_signing_key(
            self.auth_schema,
            credentials.secret_access_key,
            to_signer_date(date),
            partition,
            service_name,
        )
        signature = _generate_signature(signing_key, string_to_sign)
        authorization = self._generate_authorization(
            self.auth_type,
            credentials.access_key_id,
            scope,
            signed_headers,
            signature,
        )
        headers["Authorization"] = authorization
        return headers


class Validator:
    @staticmethod
    def _parse_authorization(authorization: str) -> typing.Tuple[str, str, str, str]:
        auth_type, _, credentials = authorization.partition(" ")
        parts = credentials.split(", ")
        data = {}
        for part in parts:
            k, _, v = part.partition("=")
            data[k.lower()] = v

        return auth_type, data["credential"], data["signedheaders"], data["signature"]

    @staticmethod
    def _parse_credential(credential: str) -> typing.Tuple[str, str]:
        return tuple(credential.split("/", maxsplit=1))

    @staticmethod
    def _parse_scope(scope: str) -> typing.Tuple[str, str, str]:
        """Convert a date/partition/service_name/request_tyep scope into a (date, partition, service_name) tuple."""
        return tuple(scope.split("/")[:-1])

    @staticmethod
    def _parse_key_date(headers: multidict.CIMultiDict) -> str:
        """Convert a date/partition/service_name/request_tyep scope into a (date, partition, service_name) tuple."""
        key_date = headers.get("x-amz-date", headers.get("x-neos-date"))
        if key_date is None:
            msg = "Missing supported date header"
            raise error.SignatureError(msg)
        # TODO: detect drift/replay

        return key_date

    @staticmethod
    def _generate_canonical_request_hash(
        method: str,
        url: httpx.URL,
        headers: multidict.CIMultiDict,
        signed_headers: str,
        content_sha256: str,
    ) -> str:
        """Get canonical request hash."""
        canonical_headers = _recreate_canonical_headers(headers, signed_headers)
        canonical_query_string = _generate_canonical_query_string(url.query)

        path = _quote(url.path or "/")
        # CanonicalRequest =
        #   HTTPRequestMethod + '\n' +
        #   CanonicalURI + '\n' +
        #   CanonicalQueryString + '\n' +
        #   CanonicalHeaders + '\n\n' +
        #   SignedHeaders + '\n' +
        #   HexEncode(Hash(RequestPayload))  noqa: ERA001
        canonical_request = (
            f"{method}\n"
            f"{path}\n"
            f"{canonical_query_string}\n"
            f"{canonical_headers}\n\n"
            f"{signed_headers}\n"
            f"{content_sha256}"
        )
        logger.debug(f"challenge canonical_request: {canonical_request}")
        return sha256_hash(canonical_request)

    @staticmethod
    def _generate_string_to_sign(auth_type: str, date: str, scope: str, canonical_request_hash: str) -> str:
        """Get string-to-sign."""
        return f"{auth_type}\n{date}\n{scope}\n{canonical_request_hash}"

    def challenge_v4(
        self,
        method: str,
        url: httpx.URL,
        headers: multidict.CIMultiDict,
        content: typing.Tuple[bytes, None],
    ) -> typing.Tuple[str, str, str, str, str]:
        logger.debug(f"url: {url}")
        logger.debug(f"headers: {headers}")
        auth_type, credential, signed_headers, signature = self._parse_authorization(headers["Authorization"])

        auth_schema = auth_type.split("-")[0]
        header_prefix = "x-neos" if auth_schema == "NEOS4" else "x-amz"
        content_header = f"{header_prefix}-content-sha256"
        content_sha256 = (
            sha256_hash(content)
            if headers.get(content_header, "UNSIGNED-PAYLOAD") != "UNSIGNED-PAYLOAD"
            else "UNSIGNED-PAYLOAD"
        )

        access_key, scope = self._parse_credential(credential)
        date, partition, service_name = self._parse_scope(scope)
        key_date = self._parse_key_date(headers)

        canonical_request_hash = self._generate_canonical_request_hash(
            method,
            url,
            headers,
            signed_headers,
            content_sha256,
        )
        string_to_sign = self._generate_string_to_sign(auth_type, key_date, scope, canonical_request_hash)
        logger.debug(f"challenge scope: {scope}")
        logger.debug(f"challenge string_to_sign: {string_to_sign}")

        return auth_type, access_key, scope, string_to_sign, signature

    def validate_challenge(
        self,
        auth_schema: str,
        scope: str,
        string_to_sign: str,
        signature: str,
        secret_key: str,
    ) -> bool:
        date, partition, service_name = self._parse_scope(scope)
        signing_key = _generate_signing_key(
            auth_schema,
            secret_key,
            date,
            partition,
            service_name,
        )
        signature_ = _generate_signature(signing_key, string_to_sign)

        if signature_ != signature:
            logger.info(f"Provided signature: {signature}, Calculated signature: {signature_}")
            msg = "Invalid signature"
            raise error.SignatureError(msg)

        return True



import abc
import re
import typing
from dataclasses import dataclass
from enum import Enum

import pydantic
import pydantic_settings

from neos_common import error


class Config(pydantic_settings.BaseSettings):
    """Service configuration base class."""

    name: str = "common"

    logging_level: str = pydantic.Field("INFO", alias="NC_LOGGING_LEVEL")

    # -- DATABASE --

    postgres_host: str = pydantic.Field(..., alias="NC_POSTGRES_HOST")
    postgres_user: str = pydantic.Field(..., alias="NC_POSTGRES_USER")
    postgres_database: str = pydantic.Field(..., alias="NC_POSTGRES_DATABASE")
    postgres_password: str = pydantic.Field(..., alias="NC_POSTGRES_PASSWORD")
    postgres_port: int = pydantic.Field(5432, alias="NC_POSTGRES_PORT")

    postgres_pool_min_size: int = pydantic.Field(1, alias="NC_POSTGRES_POOL_MIN_SIZE")
    postgres_pool_max_size: int = pydantic.Field(1, alias="NC_POSTGRES_POOL_MAX_SIZE")

    @property
    def postgres_dsn(self) -> str:
        """Generate postgres dsn from provided configuration."""
        return "postgresql://{}:{}@{}:{}/{}".format(
            self.postgres_user,
            self.postgres_password,
            self.postgres_host,
            self.postgres_port,
            self.postgres_database,
        )

    @property
    def logging_config(self) -> dict:
        """Generate default logging config."""
        return {
            "version": 1,
            "disable_existing_loggers": True,
            "formatters": {
                "default": {
                    "format": "[%(asctime)s][%(name)s][%(levelname)s]: %(message)s",
                    "datefmt": "%Y-%m-%d %H:%M:%S",
                },
            },
            "handlers": {
                "default": {
                    "class": "logging.StreamHandler",
                    "level": self.logging_level.upper(),
                    "formatter": "default",
                },
            },
            "loggers": {
                self.name: {
                    "handlers": ["default"],
                    "level": self.logging_level.upper(),
                    "propagate": False,
                },
                "yoyo.migrations": {
                    "handlers": ["default"],
                    "level": "INFO",
                    "propagate": False,
                },
                "neos_common": {
                    "handlers": ["default"],
                    "level": self.logging_level.upper(),
                    "propagate": False,
                },
            },
        }


class ApiConfig(Config):
    """API Service configuration base class."""

    name: str = "common-api"

    raw_api_prefix: str = pydantic.Field("", alias="NC_API_PREFIX")

    @property
    def api_prefix(self) -> str:
        """Ensure api prefix starts with /."""
        api_prefix = self.raw_api_prefix
        if api_prefix and not api_prefix.startswith("/"):
            api_prefix = f"/{api_prefix}"
        return api_prefix

    # -- KEYCLOAK --

    keycloak_host: str = pydantic.Field(..., alias="NC_KEYCLOAK_HOST")
    keycloak_realm: str = pydantic.Field(..., alias="NC_KEYCLOAK_REALM")
    keycloak_client_id: str = pydantic.Field(..., alias="NC_KEYCLOAK_CLIENT_ID")
    keycloak_client_secret: str = pydantic.Field(..., alias="NC_KEYCLOAK_CLIENT_SECRET")

    # -- ACCESS SECRET --
    access_key_id: str = pydantic.Field("access_key_id", alias="NC_ACCESS_KEY_ID")
    secret_access_key: str = pydantic.Field("secret_access_key", alias="NC_SECRET_ACCESS_KEY")
    partition: str = pydantic.Field("ksa", alias="NC_PARTITION")

    @property
    def logging_config(self) -> dict:
        """Generate default logging config including api loggers."""
        return {
            "version": 1,
            "disable_existing_loggers": True,
            "formatters": {
                "default": {
                    "format": "[%(asctime)s][%(name)s][%(levelname)s]: %(message)s",
                    "datefmt": "%Y-%m-%d %H:%M:%S",
                },
            },
            "handlers": {
                "default": {
                    "class": "logging.StreamHandler",
                    "level": self.logging_level.upper(),
                    "formatter": "default",
                },
            },
            "loggers": {
                self.name: {
                    "handlers": ["default"],
                    "level": self.logging_level.upper(),
                    "propagate": False,
                },
                "web_error": {
                    "handlers": ["default"],
                    "level": self.logging_level.upper(),
                    "propagate": False,
                },
                "uvicorn": {
                    "handlers": ["default"],
                    "level": "INFO",
                    "propagate": False,
                },
                "yoyo.migrations": {
                    "handlers": ["default"],
                    "level": "INFO",
                    "propagate": False,
                },
                "neos_common": {
                    "handlers": ["default"],
                    "level": self.logging_level.upper(),
                    "propagate": False,
                },
            },
        }


class ActionBase(Enum):
    """Action class base.

    When implementing IAM actions in a service, extend this class.
    """


class EffectEnum(Enum):
    """Default effect enum for use with IAM actions."""

    allow = "allow"
    deny = "deny"


RL = typing.TypeVar("RL", bound="ResourceLike")


@dataclass
class ResourceLike:
    all_: typing.Union[str, None] = None

    xrn: str = "urn"
    partition: str = ""
    service: str = ""
    identifier: str = ""
    account_id: str = ""
    resource_type: str = ""
    sub_type: str = ""
    resource_id: typing.Union[str, None] = None

    PATTERN_RULE = "[a-zA-Z0-9_\\-]{1,50}"
    OPTIONAL_PATTERN_RULE = "[a-zA-Z0-9_\\-]{0,50}"
    SUB_RESOURCE_TYPE_RULE = "[a-z-]{1,50}"
    ADDITIONAL_PATTERN_RULE = r"(\*|[a-zA-Z0-9_\-]{1,50})"
    # Valid are:
    # - *
    # - urn or nrn
    # - urn:partition:service:identifier:account_id:resource_type
    # - urn:partition:service:identifier:account_id:resource_type:*
    # - urn:partition:service:identifier:account_id:resource_type:resource_id
    # - urn:partition:service:identifier:account_id:resource_type:sub-type:resource_id
    #
    # Invalid are:
    # Pathlike
    # - urn:partition:service:identifier:account_id:resource_type/resource_id
    # - urn:partition:service:identifier:account_id:resource_type:resource_id/sub/path
    # - urn:partition:service:identifier:account_id:resource_type:resource_id:sub:path
    RESOURCE_PATTERN = r"^((?P<all_>\*)|(?P<xrn>[un]rn):(?P<partition>{rule}):(?P<service>{rule}):(?P<identifier>{optional_rule})?:(?P<account_id>{rule})?:(?P<resource_type>{rule})(([:](?P<sub_type>{sub_rule}))?[:](?P<resource_id>{additional_rule}))?)$".format(  # noqa: E501
        rule=PATTERN_RULE,
        optional_rule=OPTIONAL_PATTERN_RULE,
        sub_rule=SUB_RESOURCE_TYPE_RULE,
        additional_rule=ADDITIONAL_PATTERN_RULE,
    )

    @classmethod
    def parse(cls: typing.Type[RL], s: str) -> RL:
        m = re.match(cls.RESOURCE_PATTERN, s)
        if not m:
            msg = f"Could not parse the resource {s}"
            raise error.InvalidResourceFormatError(msg)
        m_dict = m.groupdict()
        return cls(
            all_=m_dict["all_"],
            xrn=m_dict["xrn"] or "",
            partition=m_dict["partition"] or "",
            service=m_dict["service"] or "",
            identifier=m_dict["identifier"] or "",
            account_id=m_dict["account_id"] or "",
            resource_type=m_dict["resource_type"] or "",
            sub_type=m_dict["sub_type"] or "",
            resource_id=m_dict["resource_id"],
        )

    @property
    def urn(self) -> str:
        if self.all_:
            return self.all_

        full_xrn = "{0.xrn}:{0.partition}:{0.service}:{0.identifier}:{0.account_id}:{0.resource_type}".format(self)
        if self.resource_id:
            resource_id = self.resource_id if not self.sub_type else f"{self.sub_type}:{self.resource_id}"
            full_xrn = f"{full_xrn}:{resource_id}"
        return full_xrn


RR = typing.TypeVar("RR", bound="ResourceReader")


@dataclass
class ResourceReader(ResourceLike):
    STAR = "*"

    @classmethod
    def to_root(cls: typing.Type[RR], other: RR) -> RR:
        return cls(
            all_=other.all_,
            xrn=other.xrn,
            partition=other.partition,
            service=other.service,
            identifier=other.identifier,
            account_id="root",
            resource_type=other.resource_type,
            sub_type=other.sub_type,
            resource_id=other.resource_id,
        )

    def to_list(self) -> typing.List[str]:
        if self.all_ is not None and self.is_any():
            return [self.all_]

        ret = [
            el or ""
            for el in [
                self.xrn,
                self.partition,
                self.service,
                self.identifier,
                self.account_id,
                self.resource_type,
                self.sub_type,
                self.resource_id,
            ]
        ]
        if not self.sub_type:
            ret.pop(6)

        return ret

    def is_any(self) -> bool:
        return self.all_ is not None and self.all_ == self.STAR

    def is_any_resource_id(self) -> bool:
        if self.resource_id == self.STAR:
            return True
        return False

    def __getitem__(self, item: typing.Union[slice, int]) -> str:
        """Get a specific item or subset of items from a resource.

        Examples:
        -------
        Resource("a:b:c:d")[0] -> "a"
        Resource("a:b:c:d")[1:3] -> "b:c"
        """
        if isinstance(item, slice):
            if item.stop == -1 and self.sub_type:
                item = slice(item.start, -2)
            return ":".join(self.to_list()[item]).rstrip(":")
        return self.to_list()[item]

    def __len__(self) -> int:
        """Return length of resource elements."""
        return len(self.to_list())


RB = typing.TypeVar("RB", bound="ResourceBase")


@dataclass
class ResourceBase(abc.ABC, ResourceLike):
    """Resource class contains information about resource.

    Args:
    ----
        partition (str): geographic location of the system
        service (str): name of the service, for the core-gateway it is fixed to the core
        identifier (typing.Union[str, None]): identifier of the core
        account_id (typing.Union[str, None]): owner id of the core
        resource_type (str): name of the resource type
        resource_id (typing.Union[str, None]): identificator of the resource, can have a nested structure
    """

    @classmethod
    def generate(
        cls: typing.Type[RB],
        xrn: str = "urn",
        partition: str = "partition",
        service: str = "service",
        identifier: str = "identifier",
        account_id: str = "owner",
        resource_type: str = "resource_type",
        sub_type: str = "",
        resource_id: typing.Union[str, tuple, None] = None,
        all_: typing.Union[str, None] = None,
    ) -> RB:
        """Generate Resource."""
        init_kwargs = {
            "xrn": xrn,
            "partition": partition,
            "service": service,
            "identifier": identifier or "",
            "account_id": account_id or "",
            "resource_type": resource_type,
            "sub_type": sub_type or "",
            "all_": all_,
        }

        if isinstance(resource_id, str):
            init_kwargs["resource_id"] = resource_id
        elif isinstance(resource_id, tuple):
            init_kwargs["resource_id"] = cls.format_resource_id(*resource_id)

        return cls(**init_kwargs)

    @property
    def urn_template(self) -> str:
        if self.all_:
            return self.all_

        full_xrn = "{0.xrn}:{0.partition}:{0.service}:{0.identifier}:{0.account_id}:{0.resource_type}".format(self)
        if self.resource_id:
            resource_id = "{resource_id}" if not self.sub_type else f"{self.sub_type}:{{resource_id}}"
            full_xrn = f"{full_xrn}:{resource_id}"
        return full_xrn

    @classmethod
    @abc.abstractmethod
    def get_resource_id_template(cls) -> str:
        pass

    @classmethod
    @abc.abstractmethod
    def format_resource_id(cls, *args) -> str:
        pass



from neos_common.client.iam_client import IAMClient  # noqa: F401
from neos_common.client.keycloak_client import KeycloakClient  # noqa: F401
from neos_common.client.registry_client import RegistryClient  # noqa: F401



import logging
import typing

from neos_common import base, schema
from neos_common.base import ResourceLike

logger = logging.getLogger(__name__)


STAR = "*"


"""
+--------+-----------------------+----------------------+---------+
| case # | statement resource_id | endpoint resource_id | outcome |
+--------+-----------------------+----------------------+---------+
|     33 | exact                 | exact                | allow   |
|     32 | exact                 | *                    | allow   |
|     31 | exact                 | null                 | deny    |
|     23 | *                     | exact                | allow   |
|     22 | *                     | *                    | allow   |
|     21 | *                     | null                 | deny    |
|     13 | null                  | exact                | deny    |
|     12 | null                  | *                    | deny    |
|     11 | null                  | null                 | allow   |
+--------+-----------------------+----------------------+---------+

TODO: rethink the case 12 with product owners
    UPDATE: case 32 deny -> allow (to support listing of data products which are allowed for user to see)
"""

_ACCESS_MAP = {
    "33": True,
    "32": True,
    "23": True,
    "22": True,
    "11": True,
}


def _map_resource_id(resource_id: typing.Union[str, None]) -> str:
    ret = 3
    if resource_id is None:
        ret = 1
    elif resource_id == STAR:
        ret = 2
    return str(ret)


def check_resource(statement: schema.Statement, resource: ResourceLike) -> bool:
    checks: typing.List[bool] = []
    if not statement.resource:
        return False

    for statement_resource_str in statement.resource:
        # Allow all resources if *
        if statement_resource_str == STAR:
            checks.append(True)
            continue

        # Validate *rn string
        statement_resource = resource.parse(statement_resource_str)

        # Exact match of the prefix parts
        prefix_check = all(
            [
                resource.partition == statement_resource.partition,
                resource.service == statement_resource.service,
                resource.identifier == statement_resource.identifier,
                (resource.account_id == statement_resource.account_id or statement_resource.account_id == "root"),
                resource.resource_type == statement_resource.resource_type,
            ],
        )

        access_case = f"{_map_resource_id(statement_resource.resource_id)}{_map_resource_id(resource.resource_id)}"
        postfix_check = _ACCESS_MAP.get(access_case, False)

        # Exacts must be equal to pass
        if access_case == "33" and statement_resource.resource_id != resource.resource_id:
            postfix_check = False

        checks.append(all([prefix_check, postfix_check]))

    return any(checks)


def _intersect_principals(statement: schema.Statement, principal_ids: typing.List[str]) -> typing.List[str]:
    intersect: typing.List[str] = list(set(principal_ids).intersection(set(statement.principal)))
    return intersect


def check_principals_in_statement(
    statement: schema.Statement,
    principals: typing.Union[schema.Principals, typing.List[str]],
) -> bool:
    """Check intersection of principal_id in principals and statement.principal list."""
    principal_ids = principals.get_principal_ids() if isinstance(principals, schema.Principals) else principals

    principal_intersection = _intersect_principals(statement, principal_ids)
    return len(principal_intersection) > 0


def get_statement_priority(
    statement: schema.Statement,
    principals: typing.Union[schema.Principals, typing.List[str]],
) -> int:
    """Determine priority.

    Biggest priority wins.

      - User is stronger than group.
      - Deny is stronger than allow.


    No priority_type        -> priority 1
    Group priority_type     -> priority 10
    User priority_type      -> priority 20
    Deny effect             -> priority +5

    Renders to
    +----------------+--------+----------+
    | principal_type | effect | priority |
    +----------------+--------+----------+
    | none           | allow  |        1 |
    | none           | deny   |        6 |
    | group          | allow  |       10 |
    | group          | deny   |       15 |
    | user           | allow  |       20 |
    | user           | deny   |       25 |
    +----------------+--------+----------+
    """
    # Priority maps
    effect_bonus = 0
    if statement.effect == base.EffectEnum.deny.value:
        effect_bonus = 5

    if not isinstance(principals, schema.Principals):
        return 1 + effect_bonus

    principal_type_priority_map = {
        schema.PrincipalType.user: 20,
        schema.PrincipalType.group: 10,
    }

    # Determine principals priority
    principals_priority = {p.principal_id: principal_type_priority_map[p.principal_type] for p in principals.principals}

    # Determine statement principals priority
    principal_ids = principals.get_principal_ids()
    principal_intersection = _intersect_principals(statement, principal_ids)

    principal_intersection_priority = [
        principals_priority[principal_id] + effect_bonus for principal_id in principal_intersection
    ]

    return sorted(principal_intersection_priority).pop()


def filter_by_principals(
    statements: schema.Statements,
    principals: typing.Union[schema.Principals, typing.List[str]],
) -> schema.Statements:
    """Filter out those principals which are not statements."""
    filtered_statements = []
    for statement in statements.statements:
        if check_principals_in_statement(statement, principals):
            filtered_statements.append(statement.model_dump())

    return schema.Statements(statements=filtered_statements)


def filter_by_action(
    statements: schema.Statements,
    action: typing.Union[base.ActionBase, str],
) -> schema.Statements:
    """Filter out those which are not * or don't match action."""
    filtered_statements = []
    for statement in statements.statements:
        action_str: str = action.value if isinstance(action, base.ActionBase) else action
        if action_str in statement.action or STAR in statement.action:
            filtered_statements.append(statement.model_dump())

    return schema.Statements(statements=filtered_statements)


def filter_by_resource(
    statements: schema.Statements,
    resource: ResourceLike,
) -> schema.Statements:
    """Filter out those which don't pass check_resource()."""
    filtered_statements = []
    for statement in statements.statements:
        if check_resource(statement, resource):
            filtered_statements.append(statement.model_dump())

    return schema.Statements(statements=filtered_statements)


def order_by_priority_ascending(
    statements: schema.Statements,
    principals: typing.Union[schema.Principals, typing.List[str]],
) -> schema.PriorityStatements:
    """Order ascending statements by priority."""
    priority_statements = []
    for statement in statements.statements:
        priority = get_statement_priority(statement, principals)
        priority_statement = schema.PriorityStatement(priority=priority, **statement.model_dump())
        priority_statements.append(priority_statement)

    priority_statements.sort(key=lambda s: s.priority)

    return schema.PriorityStatements(statements=priority_statements)
